import numpy as np
from itertools import cycle
from leymosun.stats import pdf, bootstrap_observed_matrix_ci


def apply_pbc(lst: list, upper_bound: int):
    """Apply periodic boundaries to a list, cyclic to upper_bound.

    Given list repead the list up to a upper bound.
    This corresponds to a Periodic
    Boundary Condition (PBC), turn list ot cyclic up to an upper_bound,
    that has to be greater than the list.

    Args:
        lst: list of numbers, but would work for any.
        upper_bound: lenght of the new list, that original
        list will be applied PBC

    Returns:
        Boolean, if all are real.

    """
    pool = cycle(lst)
    c = 1
    lst_period = []
    for item in pool:
        c = c + 1
        lst_period.append(item)
        if c > upper_bound:
            break
    return lst_period


def is_imaginary_zero(ee: np.array, epsilon: float = 1e-9):
    """Check if all eigenvalues are real

    Args:
        ee: vector of eigenvalues

    Returns:
        Boolean, if all are real.

    """
    ee_img = np.imag(ee)
    return all([imag < epsilon for imag in ee_img])


def empirical_spectral_density(
    ensemble_sample: list,
    mmes_order: int = -1,
    locations: np.array = np.arange(-2.05, 2.1, 0.05),
    scale: str = "no",
    epsilon: float = 1e-9,
):
    """Compute Empirical SD: Eigenvalues, densities and location of all matrices

    Args:
        ensemble_sample: List of matrices, representative of a matrix ensemble.
        mmes_order: If the ensemble is from mixed, put the max order used.
                     Defaults to -1, not mixed order. When in use, this will
                     pad the eigenvalues.
                     MMES algorithm is based on Mixed Matrix Ensemble Sampling (MMES) 
                     Suzen 2021 https://hal.science/hal-03464130
        locations: Eigenvalue locations that we compute the density.
                   defaults np.arange(-2.05, 2.1, 0.05). Note that
                   these are the bin centres.
        scale: Scale resulting eigenvalues,
               * Defaults to "no" : No scaling
               * 'wigner': Scales with sqrt(N/2), assume a square matrix
                 ensemble, not a mixed ensemble.
        epsilon: Imaginary numbers epsilon, below considered zero.

    Returns:
        The value of eigenvalues (2D), densities (2D), eigenvalue locations.

    """
    eigenvalues = []
    densities = []
    for matrix_member in ensemble_sample:
        _eigens_member = np.linalg.eigvals(matrix_member)
        if mmes_order > 0:
            _eigens_member = np.array(apply_pbc(list(_eigens_member), mmes_order))
        if is_imaginary_zero(_eigens_member, epsilon=epsilon):
            _eigens_member = np.real(_eigens_member)
        else:
            raise ValueError("Non-zero imaginary in the eigenvalues.")
        if scale == "wigner":
            N = _eigens_member.shape[0]
            _eigens_member = _eigens_member / np.sqrt(N / 2.0)
        eigenvalues.append(_eigens_member)
        _empirical_density, _locations = pdf(_eigens_member, locations)
        densities.append(_empirical_density)
    return np.array(eigenvalues), np.array(densities), _locations


def eigenvalue_on_polynomial(eigen: float, coefficients: np.array):
    """Compute transformed eigenvalue on a polynomial

    Polynomial of degree n-1 with coefficients c defined as
    c[0] e^n-1 + c[1] r^n-2 + ... + c[n]

    Args:
        eigen: Single eigen valur
        coefficients: polynomial coefficient

    Returns:
        Transformed eigenvalur

    """
    degree = len(coefficients) - 1
    new_eigen = 0
    for i in range(degree + 1):
        current_degree = degree - i
        new_eigen = new_eigen + coefficients[i] * (eigen**current_degree)
    return new_eigen


def unfold_spectra(eigenvalues: np.array, iqr: bool = True, deg_max: int = 32):
    """Unfold given spectra and compute nearest neigbour spacing

    Using polynomial fits, with optimal degree finding that
    makes mean spacing 1.0, the closest. Then transform
    them to new values.

    We remove outliers as suggested in the literature.

    - O. Bohigas, in Random matrices and chaotic dynamics
    (LesHouches Session LII, North-Holland, 1989).
    - On the spectral unfolding of chaotic and mixed systems,
    Abuelenin, Physica A, 492, 564-570 (2018)

    Args:
        eigenvalues: Spectra
        iqr: Only consider within inter quantile range, defaults to True.

    Returns:
        Tuple of 3:
        Transformed eigenvalues (unfolded spectra).
        nearest neigbour spacings of unfolded spectra and the degree
        of the optimal polynomial

    """
    if iqr:
        q25 = np.quantile(eigenvalues, 0.25)
        q75 = np.quantile(eigenvalues, 0.75)
        eigenvalues = eigenvalues[np.where(eigenvalues >= q25)[0]]
        eigenvalues = eigenvalues[np.where(eigenvalues <= q75)[0]]
    eigenvalues_sorted = np.sort(eigenvalues)
    x = eigenvalues_sorted
    N = len(eigenvalues_sorted)
    y = np.arange(N)
    degrees = np.arange(1, deg_max)
    fluctuation_levels = []
    for deg in degrees:
        coefficients = np.polyfit(x, y, deg=deg)
        x_new = [eigenvalue_on_polynomial(x, coefficients) for x in eigenvalues_sorted]
        fluctuation_levels.append(np.mean(np.diff(x_new)))
    degree_ix = np.argmin(np.abs(np.array(fluctuation_levels) - 1.0))
    deg_opt = degrees[degree_ix]
    coefficients = np.polyfit(x, y, deg=deg_opt)
    unfolded_eigenvalues = np.array(
        [eigenvalue_on_polynomial(x, coefficients) for x in eigenvalues_sorted]
    )
    unfolded_eigen = np.sort(unfolded_eigenvalues)
    return unfolded_eigen, np.diff(unfolded_eigen), deg_opt


def nnsd(eigenvalues: np.array, locations: np.array = np.arange(0.0, 5.0, 0.1)):
    """Compute Nearest-Neigbour Spacing Densities (NNSD) given eigenvalues with 
       polynomial unfolding.

    Args:
        eigenvalues: 2D, eigenvalues over repeaded samples.
        locations: Centers to compute PDF of NNSD.

    Returns:
        Tuple of two: NNSDs 2D and location bins

    """
    nnsd_densities = []
    for eigenvalue in eigenvalues:
        _, ueigenvalues_nn, _ = unfold_spectra(
            eigenvalue, deg_max=30, iqr=True
        )
        density, _locations = pdf(ueigenvalues_nn, locations=locations)
        nnsd_densities.append(density)
    return np.array(nnsd_densities), np.array(_locations)



import numpy as np
from typing import Callable
from leymosun.random import binomial


def ensemble(
    matrix_order: int, ensemble_size: int, sampler: Callable[[int], np.array]
) -> list:
    """Generate a representative matrix ensemble given sampler function

    Args:
        matrix_order: A square matrix size, order.
        ensemble_size: The number of matrices to generate.
        sampler: A function that samples the matrix ensemble.

    Returns:
        Sampled representative matrix ensemble, 1D list.
    """
    return [sampler(matrix_order) for _ in range(ensemble_size)]


def mixed_ensemble(
    matrix_order: int,
    ensemble_size: int,
    degree_of_mixture: float,
    sampler: Callable[[int], np.array],
) -> list:
    """Generate a representative mixed matrix ensemble given sampler function.
       Implements Mixed Matrix Ensemble Sampling (MMES) algorithm [suzen21]
       Suzen 2021 https://hal.science/hal-03464130

    Args:
        matrix_order: A square matrix size, order.
        ensemble_size: The number of matrices to generate.
        degree_of_mixture: This sets the mixture level [0,1],
        closer to 1.0 ensemble is closer to
        sampler: A function that samples the matrix ensemble.

    Returns:
        Sampled representative mixed matrix ensemble, 1D list.
    """
    mixed_sample = []
    for _ in range(ensemble_size):
        mixed_order = binomial(n=matrix_order, p=degree_of_mixture)
        mixed_sample.append(sampler(mixed_order))
    return mixed_sample


def offdiagonal(A: np.array) -> np.array:
    """Extract offdiagonal elements from a square matrix into a vector

    Args:
        A: A square matrix

    Returns:
        np.array: 1D array.
    """
    matrix_order = A.shape[0]
    tril_ix = np.tril_indices(matrix_order, k=-1)
    triu_ix = np.triu_indices(matrix_order, k=1)
    od = A[tril_ix].flatten()
    return np.hstack([od, A[triu_ix].flatten()]).flatten()



from leymosun.spectral import (
    apply_pbc,
    is_imaginary_zero,
    empirical_spectral_density,
    eigenvalue_on_polynomial,
    unfold_spectra,
    nnsd
)
from leymosun.matrix import ensemble, mixed_ensemble
from leymosun.gaussian import goe
from leymosun.random import normal
import numpy as np


def test_apply_pbc():
    v = [1, 2, 3, 4]
    v_pbc = apply_pbc(v, 10)
    assert len(v_pbc) == 10
    for element, i in zip([1, 2, 3, 4, 1, 2, 3, 4, 1, 2], range(10)):
        assert element == v_pbc[i]


def test_is_imaginary_zero():
    vec = np.array([1.1 + 1e-10j, 1.2 + 1e-10j, 1.3 + 1e-10j])
    assert is_imaginary_zero(vec)


def test_empirical_spectral_density_goe():
    matrices = ensemble(matrix_order=100, ensemble_size=40, sampler=goe)
    _, density, locations = empirical_spectral_density(matrices, scale="wigner")
    mean_inx = np.argmax(density[0])
    mean = locations[mean_inx]
    assert np.abs(mean) > 0.0


def test_empirical_spectral_density_mixed_goe():
    matrices = mixed_ensemble(
        matrix_order=100, ensemble_size=30, degree_of_mixture=0.8, sampler=goe
    )
    _, density, locations = empirical_spectral_density(matrices, scale="wigner", mmes_order=100)
    mode_inx = np.argmax(density[0])
    mode = locations[mode_inx]
    assert np.abs(mode) > 1.0


def test_eigenvalue_on_polynomial():
    x = 0.1
    coeff = np.array([2.1, 1.1, 2.2, 1.3])
    x_expected = 1.5331000000000001
    x_new = eigenvalue_on_polynomial(x, coeff)
    assert np.abs(x_new - x_expected) < 1e-6


def test_unfold_spectra():
    vec, _, deg = unfold_spectra(np.arange(1000), deg_max=4)
    assert np.sum(vec) > 124700
    assert deg == 1
    
def test_nnsd():
    sim_e2 = np.array([normal(size=1000) for _ in range(10)])
    nnsd_densities, locations = nnsd(sim_e2)
    assert nnsd_densities.shape[1] == 48



from numpy.random import PCG64, RandomState, BitGenerator
import secrets
import numpy as np


def he_random_integer(bits: int = 64):
    """Generate high-entropy random integer given bits

    Generate a random integer given bits, that is suitable
    for random seeding in simulations.

    Args:
        bits: Number of bits to use, defaults to 64

    Returns:
        Random integer in given bits.

    """
    return secrets.randbits(bits)


def get_rng(bitgen: BitGenerator = PCG64, seed_bits: int = 64):
    """Get new RNG state

    Generate a new RNG state with a 'random' seed
    from the pool

    Args:
        bitgen: Bitgenerator, defaults PCG64
        seed_bits: Number of bits to use to generate seed.

    Returns:
        NumPy random state

    """
    return RandomState(bitgen(seed=he_random_integer(bits=seed_bits)))


def binomial(n: int, p: float, bitgen: BitGenerator = PCG64, seed_bits: int = 64):
    """Get a binomial random number with strong randomness

     Args:
        n: Number of trials, or max integer n, defining [1,n].
        p: probability of success, or probability of selecting the integer between [1,n]
        seed_bits: Number of bits to use to generate seed.
        bitgen: Bitgenerator, defaults PCG64
        seed_bits: Number of bits to use to generate seed.

    Returns:
        integer value from binomial distribution

    """
    rng = get_rng(bitgen=bitgen, seed_bits=seed_bits)
    return rng.binomial(n, p)


def randint(
    low: int,
    high: int,
    size: int,
    dtype: np.dtype,
    bitgen: BitGenerator = PCG64,
    seed_bits: int = 64,
):
    """Uniform strong random integer on the given interval

     Args:
        low: Lower bound [low, high]
        high: Higher bound [low, high]
        size: Number of integers to generate
        dtype: Number of bits to use to generate seed.
        bitgen: Bitgenerator, defaults PCG64
        seed_bits: Number of bits to use to generate seed.

    Returns:
        integer value from binomial distribution

    """
    rng = get_rng(bitgen=bitgen, seed_bits=seed_bits)
    return rng.randint(low, high, size, dtype)


def random(size: int, bitgen: BitGenerator = PCG64, seed_bits: int = 64):
    """Uniform strong random float in the half open interval [0.1.0]

     Args:
        size: Number of floats to generate
        bitgen: Bitgenerator, defaults PCG64
        seed_bits: Number of bits to use to generate seed.

    Returns:
        np.array of float within the half open interval [0.1.0]

    """
    rng = get_rng(bitgen=bitgen, seed_bits=seed_bits)
    return rng.random(size)

def normal(loc:float=0.0, scale:float=1.0, size:int=None, bitgen: BitGenerator = PCG64, seed_bits: int = 64):
    """Generate Normal distribution numbers, or given Gaussian parameters

     Args:
        loc: mean parameter, defaults to 0.0 for normal.
        scale: sigma, standard deviation, defaults to 1.0 for normal.
        size: Number of floats to generate
        bitgen: Bitgenerator, defaults PCG64
        seed_bits: Number of bits to use to generate seed.

    Returns:
        np.array of float within the half open interval [0.1.0]

    """
    rng = get_rng(bitgen=bitgen, seed_bits=seed_bits)
    return rng.normal(loc=loc, scale=scale, size=size)

def choice(size:int, ndraws:int, replace:bool, bitgen: BitGenerator = PCG64, seed_bits: int = 64):
    """Draw integers from a sequence given size, with or without replacement.
       Zero being the first index.
    
     Args:
        size: upper size of the set.
        ndraws: Number of draws
        replace: With or without replacement
        bitgen: Bitgenerator, defaults PCG64
        seed_bits: Number of bits to use to generate seed.

    Returns:
        np.array of integers.

    """
    rng = get_rng(bitgen=bitgen, seed_bits=seed_bits)
    return rng.choice(a=size, size=ndraws, replace=replace)



import numpy as np
from leymosun.stats import pdf
from leymosun.random import normal, choice

def test_pdf():
    values = normal(0, 1.0, 1000)
    locations = np.arange(-5.0, 5.1, 0.1)
    density, locations = pdf(values, locations)
    mean_index = np.argmax(density)
    assert locations[mean_index] > -1.0
    assert locations[mean_index] < 1.0
    
def test_pdf2():
    values = normal(size=1000)
    locations_all = np.arange(-5.0, 5.0, step=1e-2)
    inx = choice(len(locations_all), 100, replace=False)
    locations = np.sort(locations_all[inx])# non-equidistant
    density, _locations = pdf(values, locations)
    len(density) == len(_locations)



import numpy as np
from leymosun.random import get_rng
from numpy.random import PCG64, BitGenerator


def goe(matrix_order: int, scale: float = 1.0, loc: float= 0.0, bitget: BitGenerator = PCG64, seed_bits: int = 64) -> np.array:
    """Sample a square matrix from Gaussian Orthogonal Ensemble (GOE)
    
     Wigner, E. P. (1955). "Characteristic Vectors of Bordered Matrices with 
     Infinite Dimensions," 
     Annals of Mathematics*, 62(3), 548–564.

    Args:
        matrix_order: Order of the square matrix
        scale: Sigma of the Gaussian, defaults to 1.0
        loc: Location of the Gaussian, defaults to 0.0
        bitgen: Bitgenerator, defaults PCG64
        seed_bits: Number of bits to use to generate seed.

    Returns:
        np.array: square numpy array
    """
    rng = get_rng(bitgen=bitget, seed_bits=seed_bits)
    A = rng.normal(size=(matrix_order, matrix_order), scale=scale, loc=loc)
    return 0.50 * (A + A.transpose())


def wigner(locations: np.array, domasigmin_boundary=2.0):
    """Wigner semi-circle density for GOE

    Density is defined as
    \rho(\lambda) = \frac{2}{\pi \cdot \sigma^{2}} \sqrt{\sigma^{2}-\lambda^{2}}

    Args:
        locations: An eigenvalue locations.
        domain_boundary: Domain boundary value (sigma for GOE), [-2sigma, 2sigma]

    Returns:
        The density at the locations, list.

    """
    domain_boundary_sqr = domasigmin_boundary**2
    return [
        2.0
        * np.sqrt(np.abs(domain_boundary_sqr - x**2))
        / (domain_boundary_sqr * np.pi)
        for x in locations
    ]


def wigner_spacing(s:float):
    """ Nearest Neigbour Spacing Distribution
    
    Wigner, E. P. (1957). "Statistical Theory of Spectra: Fluctuations" 
    Annals of Mathematics*, 65(2), 203–216.
    
    P(s) = \frac{\pi}{2} \cdot s \exp(-\frac{\pi}{4} \cdot s^{2})
    s > 0
    
    Args:
        s: Spacing value

    Returns:
        float, analytic value
    """
    return 0.5*np.pi*s*np.exp(-0.25*np.pi*s*s)



from leymosun.gaussian import goe
from leymosun.matrix import offdiagonal
import numpy as np


def test_goe_shape():
    A = goe(1000)
    assert A.shape[0] == 1000


def test_goe_mean():
    epsilon_mean = 0.1
    A = goe(1000)
    A_diag = np.diag(A)
    A_odiag = offdiagonal(A)
    assert np.abs(np.mean(A_odiag)) < epsilon_mean
    assert np.abs(np.mean(A_diag)) < epsilon_mean


def test_goe_variance():
    epsilon_variance = 0.2
    A = goe(1000)
    A_diag = np.diag(A)
    A_odiag = offdiagonal(A)
    assert np.abs(np.var(A_odiag) - 0.5) < epsilon_variance
    assert np.abs(np.var(A_diag) - 1.0) < epsilon_variance
    
def test_goe_variance_scale():
    epsilon_variance = 0.5
    A = goe(1000, scale=2.0)
    A_diag = np.diag(A)
    A_odiag = offdiagonal(A)
    assert np.abs(np.var(A_odiag) - 2.0) < epsilon_variance
    assert np.abs(np.var(A_diag) - 4.0) < epsilon_variance



from leymosun.random import he_random_integer, get_rng, binomial, normal, choice
from numpy.random import RandomState, BitGenerator, PCG64
import numpy as np


def test_random_integer():
    seed_int = he_random_integer(bits=64)
    assert seed_int.bit_length() < 65
    seed_int = he_random_integer(bits=32)
    assert seed_int.bit_length() < 33


def test_get_rng():
    rng = get_rng(bitgen=PCG64, seed_bits=42)
    assert isinstance(rng, RandomState)


def test_binomial():
    n = 1000
    p = 0.7
    values = np.array([binomial(n, p) for _ in range(1000)])
    assert np.mean(values) < 1.1 * n * p


def test_normal():
    n = 1000
    mu = 3.0
    sigma = 1.0
    values = normal(loc=mu, scale=sigma, size=n)
    assert np.mean(values) < 3.1


def test_choice():
    size = 5
    ndraws = 10
    replace = True
    values = choice(size, ndraws, replace)
    assert np.all([(x in [0, 1, 2, 3, 4, 5]) for x in values])



import numpy as np
from leymosun.random import choice


def pdf(values: np.array, locations: np.array) -> tuple[np.array, np.array]:
    """Compute density given values and arbitary monotonic locations of bin centres.

     Args:
        values: set of values that density should be computed.
        locations: locations where density is defined, these are bin centres!
                   numpy uses bin-edges, so we would lose 1st and last bin centers,

    Returns:
        density, location arrays as tuples.

    """
    delta = np.diff(locations)/2.0
    _locations_edges = locations[1:] - delta
    density, _location_edges = np.histogram(values, bins=_locations_edges, density=True)
    _locations = _location_edges + delta
    return density, _locations[:-1]


def bootstrap_observed_matrix_ci(
    observations: np.array,
    nboot: int = 1000,
    lower_quantile: float = 0.025,
    upper_quantile: float = 0.975,
):
    """Compute bootstrapped confidence intervals (CI) and the mean given experimental matrix

     Args:
        observations: Experimental matrix, row are different measurements, columns are different variates.
        nboot: Number of bootstraps to use, defaults to 1000
        lower_quantile: Lower bound for the confidence interval, defaults to 2.5%
        upper_quantile: Upper bound for the confidence interval, defaults to 97.5%

    Returns:
       Tuple of vectors, observed_mean, observed_upper, observed_lower

    """
    nrow = observations.shape[0]
    ncol = observations.shape[1]

    bootmatrix = []
    for _ in range(nboot):
        boot_indices = choice(nrow, nrow, replace=True)
        observations_boot = observations[boot_indices,]
        observed_means = np.mean(observations_boot, axis=0)
        bootmatrix.append(observed_means)
    bootmatrix = np.array(bootmatrix)

    observed_mean = np.mean(bootmatrix, axis=0)
    observed_upper = np.array(
        [np.quantile(bootmatrix[:, col], q=upper_quantile) for col in range(ncol)]
    )
    observed_lower = np.array(
        [np.quantile(bootmatrix[:, col], q=lower_quantile) for col in range(ncol)]
    )
    return observed_mean, observed_upper, observed_lower



from leymosun.matrix import ensemble, mixed_ensemble, offdiagonal
from leymosun.gaussian import goe
import numpy as np


def test_ensemble_size_goe():
    ensemble_size = 10
    matrix_order = 1000
    ensemble_sample = ensemble(matrix_order, ensemble_size, goe)
    assert len(ensemble_sample) == ensemble_size
    
def test_ensemble_orders_goe():
    ensemble_size = 10
    matrix_order = 1000
    ensemble_sample = ensemble(matrix_order, ensemble_size, goe)
    for matrix_sample in ensemble_sample:
        assert matrix_sample.shape[0] == matrix_order

def test_mixed_ensemble_size_goe():
    ensemble_size = 10
    matrix_order = 1000
    degree_of_mixture = 0.9
    ensemble_sample = mixed_ensemble(matrix_order, ensemble_size, degree_of_mixture, goe)
    assert len(ensemble_sample) == ensemble_size
    
def test_mixed_ensemble_orders_goe():
    ensemble_size = 10
    matrix_order = 1000
    degree_of_mixture = 0.9
    ensemble_sample = mixed_ensemble(matrix_order, ensemble_size, degree_of_mixture, goe)
    for matrix_sample in ensemble_sample:
        assert matrix_sample.shape[0] < matrix_order

def test_offdiagonal():
    A = np.array([[11,12,13],[21,22,23],[31,32,33]])
    A_od = offdiagonal(A)
    assert A_od.shape[0] == 6
    for off_element in [12,13,21,23,31,32]:
        assert off_element in A_od


from setuptools import setup

fp = open("./leymosun/version.py")
exec(fp.read())

with open("README.md") as f:
    long_description = f.read()

import sys

if (sys.version_info.major, sys.version_info.minor) < (3, 11):
    sys.exit(" Python <= 3.11 is not supported or preferred")

setup(
    name="leymosun",
    version=__version__,
    description="High-Entropy Randomness Research Toolkit. High-Entropy Random Number Generator (HE-RNGs). \
                Generation of random matrices in canonical ensembles. Generating of mixed matrix ensembles via \
                Mixed Matrix Ensemble Sampling (MMES) algorithm. \
                Matrix utilities, spectral analysis tools, analytic expressions from theory and \
                uncertainty quantification tools.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/msuzen/leymosun",
    author="M.Suzen",
    author_email="suzen@acm.org",
    license="GPL-3",
    packages=["leymosun"],
    install_requires=[
        "numpy == 2.3.1",
        "matplotlib == 3.10.3",
        "dill == 0.4.0"
    ],
    test_suite="tests",
    extras_require={
        "test":["pytest==8.4.2", "coverage==7.10.7", "pytest-cov==7.0.0"]
    },
    zip_safe=False,
)



from collections import namedtuple
import dill

Params = namedtuple("Params", ["degree_of_mixtures", "mixed_order", "ensemble_size"])

""" 

Params: Simulation parameters in a namedtuple with the members.

degree_of_mixtures: List of degree of mixtures to compute.  
mixed_order: (Reference) matrix order.  
ensemble_size: The number of matrices to consider.  
 
"""

MixedEnsemble = namedtuple(
    "MixedEnsemble",
    [
        "uuid",
        "degree_of_mixture",
        "matrix_order",
        "ensemble_size",
        "eigenvalues",
        "spectra_values",
        "spectra_locations",
    ],
)

""" 

MixedEnsemble: Mixed ensemble storage for spectral distributions

uuid: unique uuid4. 
degree_of_mixture: Mixture strength.  
matrix_order: Order of the (reference) matrix.  
ensemble_size: Number of matrices to consider.  
eigenvalues: All eigenvalues vector in the ensemble.  
spectra_values: ESD, Empirical Spectral Density if eigenvalues.  
spectra_locations: ESD locations, eigenvalues, bin centers in histogram parlance.  
 
"""

MixedEnsembleSet = namedtuple(
    "MixedEnsembleSet", ["degree_of_mixtures", "mixed_ensembles"]
)

""" 

MixedEnsembleSet: Mixed ensemble sets.

degree_of_mixtures: List/array of mixture strenghts.  
mixed_ensembles: List/array of MixedEnsemble stroges, corresponding mixture strengths.  

"""

SpectralDataStorage = namedtuple("SpectralDataStorage", ["sim_params", "mixed_ensemble_set"])

""" 

SpectralDataStorage: Mixed ensemble datastorage full.  

sim_params: Sim Params. 
mixed_ensemble_set: Set of mixed ensembles.  

"""

def write_storage(storage: SpectralDataStorage, filename: str):
    """Write SpectralDataStorage object to a file.
    
    Args:
        storage: SpectralDataStorage object
        filename: Filename to store, using .dill extension is recommended. 

    Returns:
        None
    
    """
    return dill.dump(storage, open(filename, "wb"))


def read_storage(filename: str) -> SpectralDataStorage:
    """Write SpectralDataStorage object to a file.
    
    Args:
        filename: Filename to read from, using .dill extension is recommended. 

    Returns:
        SpectralDataStorage object
    
    """
    return dill.load(open(filename, "rb"))
